import logging
import os
import sqlite3 as sqlite
import sqlite_spellfix
from datetime import datetime

import configargparse

log = logging.getLogger(__name__)

parser = configargparse.get_argument_parser()
parser.add_argument("--db", default="/tmp/plates.db")
parser.add_argument(
    "--db-reset", action="store_true", default=False, help="reset database"
)


class Db:
    """
    stato = 0: plate non localizzato
    stato = 1: plate localizzato (dal timestamp ottengo le coordinate GPS)
    stato = 2: plate filtrato
    stato = 4: plate visurato
    stato = -2: plate scartato perché altro migliore presente
    stato = 3: plate spedito
    stato = -3: plate da non spedire perché in zona off
    stato = 31: plate scartato per non averlo localizzato
    """

    def __init__(self, filename, reset=False):
        self.filename = filename
        self.db = None
        if reset and os.path.exists(filename):
            log.warning("db reset request: remove %s", self.filename)
            os.remove(self.filename)

        self.connect()  # solleva eccezione

    def connect(self):
        if self.db:
            return

        self.db = sqlite.connect(self.filename)  # + '.' + str(self.camera_id))
        self.db.row_factory = sqlite.Row
        self.db.enable_load_extension(True)
        try:
            self.db.load_extension(sqlite_spellfix.extension_path())
            # self.db.load_extension("spellfix")  # spellfix.so has to be in LD_LIBRARY_PATH
        except Exception:
            log.warning("no spellfix extension")
        self.db.enable_load_extension(False)
        log.info("initialize sqlite db=%s", self.filename)

        with self.db:
            self.db.execute(
                """CREATE TABLE IF NOT EXISTS plates (
                id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
                key TEXT NOT NULL,
                object_id TEXT,
                state INTEGER DEFAULT 0,
                plate_ascii TEXT,
                plate TEXT,
                score REAL DEFAULT 0,
                timestamp INTEGER DEFAULT 0,
                quarter INTEGER DEFAULT 0,
                x INTEGER DEFAULT 0,
                y INTEGER DEFAULT 0,
                w INTEGER DEFAULT 0,
                h INTEGER DEFAULT 0,
                angle INTEGER DEFAULT 0,
                parking INTEGER DEFAULT 0,
                latitude REAL DEFAULT 0,
                longitude REAL DEFAULT 0,
                elevation REAL DEFAULT 0,
                speed REAL DEFAULT 0,
                zone INTEGER DEFAULT 0,
                nationality TEXT DEFAULT 'ITA'
                );"""
            )  #! NG: patch issue 38

            self.db.execute(
                """CREATE INDEX IF NOT EXISTS plates_timestamp_index on plates(state, timestamp);"""
            )
            self.db.execute(
                """CREATE INDEX IF NOT EXISTS plates_ascii_index on plates(plate_ascii);"""
            )
            self.db.execute(
                """CREATE INDEX IF NOT EXISTS plates_score_index on plates(score);"""
            )
            self.db.execute("""CREATE INDEX IF NOT EXISTS plates_key on plates(key);""")

            self.db.execute(
                """PRAGMA journal_mode=TRUNCATE"""
            )  # On many systems, truncating a file is much faster than deleting the file since the containing directory does not need to be changed.

    def add(
        self,
        key="",
        object_id="",
        plate_ascii="",
        plate="",
        score=0.0,
        timestamp=0,
        quarter=0,
        x=0,
        y=0,
        w=0,
        h=0,
        angle=0,
        parking=0,
        latitude=0,
        longitude=0,
        elevation=0,
        speed=0,
        nationality="ITA",
    ):
        # if not (plate_ascii == "xxxEY378WA" or plate_ascii[:3]=="DW0"):
        #     return
        with self.db:
            self.db.execute(
                """INSERT INTO plates 
                (key,object_id,plate_ascii,plate,score,timestamp,quarter,x,y,w,h,angle,parking,latitude,longitude,elevation,speed,nationality) 
                VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)""",
                (
                    key,
                    object_id,
                    plate_ascii,
                    plate,
                    score,
                    timestamp,
                    quarter,
                    x,
                    y,
                    w,
                    h,
                    angle,
                    parking,
                    latitude,
                    longitude,
                    elevation,
                    speed,
                    nationality,  #! NG: patch issue 38
                ),
            )

    def get(self, state=0):
        with self.db.cursor(ContextManagerCursor) as cursor:
            cursor.execute(
                "SELECT * FROM plates WHERE state=? order by timestamp", (state,)
            )
            return cursor.fetchall()
            # async for plate in cursor:
            #    yield plate

    def getLatest(self):
        with self.db.cursor(ContextManagerCursor) as cursor:
            cursor.execute("SELECT * FROM plates ORDER BY id DESC LIMIT 1")
            return cursor.fetchone()

    def getByKey(self, key):
        with self.db.cursor(ContextManagerCursor) as cursor:
            cursor.execute("SELECT * FROM plates WHERE key=?", (key,))
            return cursor.fetchone()

    def getById(self, primary_key):
        with self.db.cursor(ContextManagerCursor) as cursor:
            cursor.execute("SELECT * FROM plates WHERE id=?", (primary_key,))
            return cursor.fetchone()

    def getOlder(self, state=0, validate_delay=10):
        timestamp = datetime.utcnow().timestamp() - validate_delay
        with self.db.cursor(ContextManagerCursor) as cursor:
            cursor.execute(
                "SELECT * FROM plates WHERE state=? and timestamp < ? order by timestamp",
                (state, timestamp),
            )
            return cursor.fetchall()

    def updateLocation(self, plate, location, state):
        self.db.execute(
            "UPDATE plates SET latitude=?,longitude=?,elevation=?,state=?,speed=? WHERE id=?",
            (
                location.latitude,
                location.longitude,
                location.elevation,
                state,
                location.speed,
                plate["id"],
            ),
        )

    def updateZone(self, plate, zone):
        self.db.execute(
            "UPDATE plates SET zone=? WHERE id=?", (int(zone or 0), plate["id"])
        )

    def updateState(self, plate, state: int):
        self.db.execute("UPDATE plates SET state=? WHERE key=?", (state, plate["key"]))

    def commit(self):
        self.db.commit()

    def rollback(self):
        self.db.rollback()

    def getSimilar(self, plate_ascii, max_distance=300, state=0):
        with self.db.cursor(ContextManagerCursor) as cursor:
            cursor.execute(
                """SELECT *, editdist3(plate_ascii, ?) as distance 
                FROM plates WHERE state=? AND distance <= ? ORDER BY score DESC;""",
                (plate_ascii, state, max_distance),
            )
            return cursor.fetchall()

    def close(self):
        self.db.close()


class ContextManagerCursor(sqlite.Cursor):
    def __enter__(self):
        return self

    def __exit__(self, a, b, c):
        self.close()


if __name__ == "__main__":
    import random
    import string

    db = Db("/tmp/plates.db")

    db.connect()

    p_ascii = "".join(random.choices(string.ascii_uppercase + string.digits, k=7))
    db.add(plate_ascii=p_ascii)

    for p in db.get():
        print("plate=", p["plate_ascii"])

    db.close()
