import aiohttp
import time
import yaml
import logging
import pkg_resources

import configargparse

parser = configargparse.get_argument_parser()
parser.add_argument("--api", default="/lab/dnn/src/mev_unit/mev_cloud.yaml")

log = logging.getLogger(__name__)

default_config = {
    "enable": True,
    "client_id": "5djZFhyPNlcR8SgL",
    "client_secret": "aRNeIPZ9rDSXqCFw3Qh",
    "extDeviceSign": "https://fes.kiunsys.com/rest_anpr/extDeviceSign",
    "recognizedPlates": "https://fes.kiunsys.com/rest_anpr/recognizedPlates",
    "refreshToken": "https://fes.kiunsys.com/rest_anpr/refreshToken",
    "timeout": 30,
    "token": "",
    "refresh_token": "",
    "token_date": None,
    "cameraCode": "galileoTest",
}


async def on_request_start(session, trace_config_ctx, params):
    log.warn("request: url='%s' headers=%s", params.url, dict(params.headers))


async def on_request_end(session, trace_config_ctx, params):
    log.warn("response: %s", await params.response.text())


class Api:
    mev_unit: str = ""
    unit_id: int = 0
    realm: str = ""

    def __init__(self, config=default_config, mev_unit="noname", debug=False):
        self.mev_unit = mev_unit
        if isinstance(config, str):
            with open(config) as f:
                self.config = yaml.safe_load(f)["api"]
        else:
            self.config = config

        self.trace_config = aiohttp.TraceConfig()
        if debug:  # logging.DEBUG >= logging.root.level:
            log.warning("enable debug")
            self.trace_config.on_request_start.append(on_request_start)
            self.trace_config.on_request_end.append(on_request_end)

        self.login: bool = False
        self.timeout = aiohttp.ClientTimeout(total=int(self.config.get("timeout", 30)))

    def __repr__(self):
        return str(self.config)

    def enabled(self):
        return self.config.get("enable")

    def hasToken(self):
        return self.config["token"] != ""

    async def extDeviceSignClient(self) -> bool:
        """
        response = {
            'token': 'c778cd06f09bd381a5b57941c58f1251099b01d6',
            'refreshToken': '200a76c37c6432e2a4dc75958a0b265ddbda13c6'
            }
        """

        request = {
            "grant_type": "password",
            "mev_unit": self.mev_unit,
            "client_secret": self.config["client_secret"],
            "lang": "en",
            "version": pkg_resources.require("mev_unit")[0].version,
        }

        async with aiohttp.ClientSession(
            timeout=self.timeout, trace_configs=[self.trace_config]
        ) as session:
            resp = await session.post(self.config["extDeviceSign"], json=request)
            if resp.status == 200:
                payload = await resp.json()
                self.login = True

                try:
                    self.realm = payload["realm"]
                    self.unit_id = payload["unit_id"]
                    self.config["token"] = payload["token"]
                    self.config["refresh_token"] = payload["refreshToken"]
                    self.config["token_date"] = time.time()
                    log.info(
                        "mev_cloud: login successfull %s@%s unit_id=%d",
                        self.mev_unit,
                        self.realm,
                        self.unit_id,
                    )
                except Exception as e:
                    log.error(str(e))
                    self.login = False
                return self.login
            if resp.status == 400:
                # {'errorCode': 0, errorDescription: 'bla bla'}
                log.warning("invalide response: %s", await resp.json())
                self.login = False
            return self.login

    async def recognizedPlatesClient(
        self, plates=[], simulate_expired=False, refreshed=1
    ) -> bool:
        """
        plate = {'plate':'xxx', latitude: 45.444, longitude: 11.444}
        """

        if refreshed > 3:  # limitiamo la chiamata ricorsiva
            log.warn(f"maximun refresh token reached {refreshed}")
            return False

        plates_cleaned = []
        for plate in plates:
            plate_cleaned = {
                "plate": plate["plate"],
                "plate_ascii": plate["plate_ascii"],
                "latitude": plate["latitude"],
                "longitude": plate["longitude"],
                "speed": plate["speed"],
                "timestamp": plate[
                    "timestamp"
                ],  # datetime.fromtimestamp(plate['timestamp']).isoformat(),
                "key": plate["key"],
                "parking": plate["parking"],
                "score": plate["score"],
            }
            plates_cleaned.append(plate_cleaned)
            log.debug("upload plate: %s", plate_cleaned)
            if "filename" in plate_cleaned:
                plate_cleaned.pop("filename")

        request = {
            "mev_unit": self.mev_unit,
            "data": plates_cleaned,
            #'readingDateTime': datetime.now().isoformat()
        }

        # if not upload:
        #     print(f"simulate connection to {self.config['recognizedPlates']} headers={{'Authorization': 'Bearer {self.config['token']}'}}")
        #     print(f"payload: {request}")
        #     return True

        async with aiohttp.ClientSession(
            timeout=self.timeout, trace_configs=[self.trace_config]
        ) as session:
            headers = (
                {"Authorization": f"Bearer {self.config['token']}"}
                if not simulate_expired
                else {}
            )
            log.debug("recognizedPlates payload: %s" % request)
            resp = await session.post(
                self.config["recognizedPlates"], json=request, headers=headers
            )
            if resp.status in (204, 200):  # esito positivo
                payload = await resp.json()
                if "items" in payload:
                    return payload["items"] == len(plates)
                return True
            if resp.status == 401:
                log.warning(
                    "access denyed: %s, try to refresh token" % await resp.json()
                )
                self.resetToken()
                if await self.refreshToken():
                    return await self.recognizedPlates(plates, refreshed=refreshed + 1)
            log.warn("upload failed with http code %s", resp.status)
            return False

    def resetToken(self):
        self.config["token"] = ""

    async def refreshToken(self) -> bool:
        """
        response = {
            'token': 'c778cd06f09bd381a5b57941c58f1251099b01d6',
            'refreshToken': '200a76c37c6432e2a4dc75958a0b265ddbda13c6'
            }

        response = {'errorCode': 0, errorDescription: 'bla bla'}
        """

        request = {
            "grant_type": "refresh_token",
            "client_id": self.config["client_id"],
            "client_secret": self.config["client_secret"],
            "refresh_token": self.config["refresh_token"],
        }
        async with aiohttp.ClientSession(
            timeout=self.timeout, trace_configs=[self.trace_config]
        ) as session:
            resp = await session.post(self.config["refreshToken"], json=request)
            if resp.status == 200:
                payload = await resp.json()
                if payload and "errorCode" in payload:
                    log.warning("TODO deve andare in resp.status=400")
                    return False
                try:
                    self.config["token"] = payload["token"]
                    self.config["refresh_token"] = payload["refreshToken"]
                    self.config["token_date"] = time.time()
                    log.debug("refreshToken response: %s" % payload)
                except Exception as e:
                    raise
                    log.error(str(e))
                    return False
                return True
            if resp.status == 400:
                log.warning("invalide response: %s" % await resp.json())
                return False
            log.warning("refreshToken response: %s" % resp)
            return False
