"""
copiare mikrotik.mib in /usr/share/snmp/mibs/MIKROTIK-MIB.txt
"""

import asyncio
import logging
import math
import os
import random
import re
import sqlite3 as sqlite
import subprocess
import sys
import time
from datetime import datetime, timezone

# import aiosqlite
import configargparse
import gpxpy
import gpxpy.gpx
import numpy as np
from gpxpy.geo import Location, LocationDelta, distance, get_course

from . import aioudp
from .gps import aiogps

# from shapely.geometry import Point

log = logging.getLogger(__name__)

parser = configargparse.get_argument_parser()
parser.add_argument("--mib", default="/usr/share/snmp/mibs/MIKROTIK-MIB.txt")
parser.add_argument("--mikrotik", default="192.168.2.31", help="mikrotik address")
parser.add_argument("--gpx-path", default="/tmp", help="gpx path to save")
parser.add_argument(
    "--gpx-autosave", default=30, type=int, help="gpx autosave frequency in seconds"
)
parser.add_argument("--gpx-db", default="/tmp/gps.db", help="gpx sqlite db")
parser.add_argument("--socket-port-gps", type=int, default=5012)
parser.add_argument("--gps", default="dummy", help="mikrotik, gpsd or file gpx to load")
parser.add_argument(
    "--gps-stop",
    type=float,
    default=0,
    help="stop car 5percent of time if 0.05. default=0",
)
parser.add_argument(
    "--db-reset", action="store_true", default=False, help="reset database"
)


class Gps:
    def __init__(self, config, filename=""):  # , no_save=False):
        self.config = config

        self.bearing_current = 0
        self.point_current = None
        self.point_previous = None
        self.counter = 0
        # self.satellites = 0

        self.gpx_saving = False
        self.update_callback = []
        self.filename = filename
        self.db = None
        self.gpx = None
        self._current_gpx_index = -1

        self.gpx_track = gpxpy.gpx.GPXTrack()
        self.time_valid = False
        self.fix = False
        self.resetGpxOut()
        self.points = []  # usata in sviluppo

        # if not loop:
        #    loop = asyncio.get_event_loop()

        # loop.create_task(self._safe_connect())

    def resetGpxOut(self):
        self.gpx_out = gpxpy.gpx.GPX()
        self.gpx_out_track = gpxpy.gpx.GPXTrack()
        self.gpx_out.tracks.append(self.gpx_out_track)
        self.filename = "{}.gpx".format(int(time.time()))
        self.gpx_out_segment = None

    def initialize(self):
        self.initializeDb()

        loop = asyncio.get_event_loop()
        if self.config.gps == "gpsd":
            loop.create_task(self.poll_aiogpsd())
            # loop.create_task(self.autoSave())
        elif self.config.gps == "dummy":
            loop.create_task(self.dummy())
        elif self.config.gps == "remote":
            pass
        else:
            loop.create_task(self.generate())

    # async def _safe_connect_db(self):
    #     try:
    #         return await self._connect()
    #     except:
    #         await self.db.close()
    #         raise

    def initializeDb(self):
        if self.config.db_reset and os.path.exists(self.config.gpx_db):
            log.warning("db reset request: remove %s", self.config.gpx_db)
            os.remove(self.config.gpx_db)

        log.info("initialize sqlite gps db %s", self.config.gpx_db)
        self.db = sqlite.connect(self.config.gpx_db)
        self.db.row_factory = sqlite.Row

        with self.db:
            self.db.execute(
                """CREATE TABLE IF NOT EXISTS locations (
                id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
                timestamp REAL DEFAULT 0,
                latitude REAL DEFAULT 0,
                longitude REAL DEFAULT 0,
                elevation REAL DEFAULT 0,
                bearing REAL DEFAULT 0,
                speed REAL DEFAULT 0,
                satellites INTEGER DEFAULT 0
                );"""
            )

            self.db.execute(
                """CREATE INDEX IF NOT EXISTS locations_timestamp_index on locations(timestamp);"""
            )

            # self.db.execute('''PRAGMA journal_mode=WAL''')

    def addCallback(self, callback):
        self.update_callback.append(callback)

    async def getLocationFromTime(self, timestamp=0, quarter=0):
        """
        cerchiamo la coppia di location p1 e p2 che contiene il timestamp dato e torniamo l'interpolazione
        """
        quarter = int(quarter)

        def _get_location_from_timestamp(_timestamp):
            with self.db.cursor(ContextManagerCursor) as cursor:
                cursor.execute(
                    """select id,latitude,longitude,elevation,bearing,timestamp,speed,(timestamp-?) as delta, min(abs(timestamp - ?)) as abs_delta from locations;""",
                    (_timestamp, _timestamp),
                )
                return cursor.fetchone()
                # print(f"select id,latitude,longitude,elevation,bearing,timestamp,(timestamp-{_timestamp}) as delta, min(abs(timestamp - {_timestamp})) as abs_delta from locations;  ----> {result}")
                # return result

        def _get_location_from_id(_id):
            with self.db.cursor(ContextManagerCursor) as cursor:
                cursor.execute(
                    """select id,latitude,longitude,elevation,bearing,timestamp,speed from locations where id=?;""",
                    (_id,),
                )
                return cursor.fetchone()

        p1 = _get_location_from_timestamp(timestamp)
        if p1["delta"] is not None:
            if p1["delta"] <= 0:
                p2 = _get_location_from_id(
                    p1["id"] + 1
                )  # se p2 è nullo dobbiamo aspettare il prossimo ciclo quando sono arrivati più dati dal GPS
            else:
                p2 = dict(p1)
                p1 = _get_location_from_id(p1["id"] - 1)
            if not p2 or not p1:
                return None

            lat = np.interp(
                timestamp,
                [p1["timestamp"], p2["timestamp"]],
                [p1["latitude"], p2["latitude"]],
            )
            lon = np.interp(
                timestamp,
                [p1["timestamp"], p2["timestamp"]],
                [p1["longitude"], p2["longitude"]],
            )
            elevation = np.interp(
                timestamp,
                [p1["timestamp"], p2["timestamp"]],
                [p1["elevation"], p2["elevation"]],
            )
            location = Location(lat, lon, elevation=elevation)
            location.speed = np.interp(
                timestamp,
                [p1["timestamp"], p2["timestamp"]],
                [p1["speed"], p2["speed"]],
            )

            angle = (
                p1["bearing"] + quarter * 90 + 45
            )  # TODO, da raffinare con il centroid
            if angle > 360:
                angle -= 360

            delta = LocationDelta(distance=4, angle=angle)
            location.move(delta)

            return location

        return None

    def checkMIB(self):
        if not os.path.exists(self.config.mib):
            print(f"MIB {self.config.mib} not present, exit.")
            sys.exit(1)
        return True

    def resetTimestamp(self):
        delta = None
        for segment in self.gpx_track.segments:
            for point in segment.points:
                if not delta:
                    delta = datetime.now(timezone.utc) - point.time
                point.time = point.time + delta

    def getBearing(self, point1, point2):
        return get_course(
            point1.latitude, point1.longitude, point2.latitude, point2.longitude
        )

    def getBearingOld(self, point1, point2):
        delta_lon = point2.longitude - point1.longitude
        y = math.sin(delta_lon) * math.cos(point2.latitude)
        x = math.cos(point1.latitude) * math.sin(point2.latitude) - math.sin(
            point1.latitude
        ) * math.cos(point2.latitude) * math.cos(delta_lon)
        brng = np.rad2deg(math.atan2(y, x))
        if brng < 0:
            brng += 360
        return brng

    def getPoints(self):
        for segment in self.gpx_track.segments:
            for point in segment.points:
                yield point

    def stopSegment(self):
        self.gpx_out_segment = None

    def close(self):
        if self.db:
            log.info("gps disconnecting db %s", self.config.gpx_db)
            self.db.close()
        self.saveGpx()

    def updateBearing(self):
        if self.point_previous and self.point_current:
            self.bearing_current = self.getBearing(
                self.point_previous, self.point_current
            )

    async def setCurrentPoint(self, point):
        self.counter += 1
        self.point_previous = self.point_current
        self.point_current = point
        self.updateBearing()
        if self.point_current:
            log.debug(
                "GPS set current location (%d): %s", self.counter, self.point_current
            )

            if self.db:
                with self.db as db:
                    db.execute(
                        """INSERT INTO locations
                        (timestamp, latitude, longitude, elevation, bearing, speed, satellites)
                        VALUES (?,?,?,?,?,?,?)""",
                        (
                            datetime.timestamp(self.point_current.time),
                            self.point_current.latitude,
                            self.point_current.longitude,
                            self.point_current.elevation,
                            self.bearing_current,
                            self.point_current.speed,
                            self.point_current.satellites,
                        ),
                    )

            for callback in self.update_callback:
                await callback(
                    p1=self.point_previous,
                    p2=self.point_current,
                    name=str(self.counter),
                )

    async def addPoint(self, point, min_speed=0):
        if self.point_current and (point.speed or 0) < min_speed:
            log.warning("skip gps point because %f < %f", point.speed, min_speed)
            return
        await self.setCurrentPoint(point)

        if not self.gpx_out_segment:
            self.gpx_out_segment = gpxpy.gpx.GPXTrackSegment()
            self.gpx_out_track.segments.append(self.gpx_out_segment)
        if self.gpx_saving:
            self.gpx_out_segment.points.append(point)

    async def track(self):
        self.resetGpxOut()
        log.info("track gpx out: %s", self.filename)
        self.gpx_saving = True
        self.saveGpx()
        # if self.gpx_out_segment:
        #    self.gpx_out_segment.points = []
        #    #self.filename = datetime.now().strftime("%Y-%m-%d_%H:%M:%S") + ".gpx"
        #    self.saveGpx()

    async def stop(self):
        self.saveGpx()
        self.gpx_saving = False
        log.info("stopped gpx tracking")

    # async def autoSave(self):
    #     if not self.config.gpx_autosave or not self.filename:
    #         return
    #     log.info("Started gps.autoSave task")
    #     while True:
    #         self.saveGpx()
    #         await asyncio.sleep(self.config.gpx_autosave)

    def saveGpx(self):
        if (
            not self.gpx_out
            or not self.filename
            or not self.gpx_saving
            or not self.gpx_out_segment
            or not self.gpx_out_segment.points
        ):
            log.warn("skip gpx out saving")
            return

        if not os.path.exists(self.config.gpx_path):
            os.makedirs(self.config.gpx_path)
        filename = os.path.join(self.config.gpx_path, os.path.basename(self.filename))
        with open(filename, "w") as f:
            f.write(self.gpx_out.to_xml())
            log.info("saved gpx %s", filename)

    async def poll_aiogpsd(self, interval=1):
        log.info("Started gpsd.poll task")
        try:
            # Example of using custom connection configuration
            async with aiogps.aiogps(
                connection_args={"host": "127.0.0.1", "port": 2947},
                connection_timeout=5,
                reconnect=0,  # do not reconnect, raise errors
                alive_opts={"rx_timeout": 5},
            ) as gpsd:
                while True:  # questo ciclo non serve
                    await asyncio.sleep(interval)
                    try:
                        # log.info(gpsd.fix.altitude)

                        # status = gpsd.fix.status

                        # if status == 0: #gps.STATUS_NO_FIX:
                        #     await self.notifyNoValidPosition()
                        #     continue
                        #
                        latest_time = None

                        async for msg in gpsd:
                            # with open("/tmp/gps.log", "a") as f:
                            #    f.write("%s\n" % dict(msg))
                            # Print received message
                            # if msg['class'] == 'SKY':
                            #     self.satellites = len([sat for sat in msg['satellites'] if sat['used']])
                            #     #print(f"satellites:{satellites} hdop:{msg['hdop']}")
                            if msg["class"] == "TPV":
                                mode = msg.get("mode", 0)
                                if (
                                    mode == 3
                                ):  #  mode=1 - no valid data, mode=2 is a 2D fix, mode=3 is a 3D fix
                                    self.fix = True
                                elif mode == 1:
                                    self.fix = False

                                if "altMSL" in msg:
                                    # print(f"{msg['lat']} {msg['lon']} {msg['altMSL']} {msg['time']} {msg['speed']} eph:{msg['eph']}")
                                    point = gpxpy.gpx.GPXTrackPoint()
                                    point.time = datetime.strptime(
                                        msg["time"], "%Y-%m-%dT%H:%M:%S.%fZ"
                                    )
                                    if (
                                        latest_time == point.time
                                    ):  # arrivano messaggi doppi, non capisco
                                        continue
                                    latest_time = point.time
                                    point.latitude = msg["lat"]
                                    point.longitude = msg["lon"]
                                    point.speed = msg["speed"]
                                    point.elevation = msg["altMSL"]
                                    point.satellites = gpsd.satellites_used
                                    point.horizontal_dilution = gpsd.hdop
                                    if point.satellites == 0:
                                        log.warning("no satellites found, skip")
                                        await self.notifyNoValidPosition()
                                        continue
                                    if not self.time_valid:
                                        if self.checkTimeSync():
                                            self.time_valid = True
                                        else:
                                            log.warning("ntp not synced")
                                    await self.addPoint(point)
                    except asyncio.IncompleteReadError:
                        log.info("Connection closed by server")
                    except asyncio.TimeoutError:
                        log.error("Timeout waiting for gpsd to respond")
                    except Exception as exc:  # pylint: disable=W0703
                        log.error(f"Error: {exc}")
        # except asyncio.CancelledError:
        #     return
        except Exception as exc:  # pylint: disable=W0703
            log.error(f"Error: {exc}")

    async def notifyNoValidPosition(self):
        log.warning("no valid position from GPS")
        point = gpxpy.gpx.GPXTrackPoint(latitude=0, longitude=0)
        point.satellites = 0
        for (
            callback
        ) in self.update_callback:  # notifichiamo comunque che non c'è la posizione
            await callback(p2=point)

    async def addPointFromRemote(self, location, timestamp):
        self.time_valid = True

        point = gpxpy.gpx.GPXTrackPoint()
        point.time = datetime.fromtimestamp(timestamp / 1000.0)
        point.latitude = location["latitude"]
        point.longitude = location["longitude"]
        point.speed = location["speed"] or 0
        point.elevation = location["altitude"] or 0
        point.satellites = 10  # gpsd.satellites_used
        point.horizontal_dilution = 1  # gpsd.hdop
        await self.addPoint(point)

    async def dummy(self):
        self.time_valid = True
        self.fix = True

        # while True:
        #     point = gpxpy.gpx.GPXTrackPoint()
        #     point.time = datetime.strptime(msg['time'], '%Y-%m-%dT%H:%M:%S.%fZ')

        #     await _local_udp.receive()  # questo arriva ogni secondo, oppure 25 frames se siamo in debug

        #     point.latitude = 0 #msg['lat']
        #     point.longitude = 0 #msg['lon']
        #     point.speed = 0 #msg['speed']
        #     point.elevation = 0 #msg['altMSL']
        #     point.satellites = 10 #gpsd.satellites_used
        #     point.horizontal_dilution = 1 #gpsd.hdop

        #     await self.addPoint(point)

    def load_gpx(self, filename):
        """
        usata durante lo sviluppo
        leggiamo un file gpx e settiamo le date dei punti a partire dal momento corrente
        """
        with open(filename) as f:
            self.gpx = gpxpy.parse(f)
        log.info(
            "loading gpx track from %s with %d points and simulate tracking",
            filename,
            self.gpx.get_points_no(),
        )
        self.gpx_track = self.gpx.tracks[0]
        self.points = []
        for segment in self.gpx_track.segments:
            self.points.extend(segment.points)

    async def generate_point(self):
        index = self._current_gpx_index + 1
        if index < len(self.points):
            point = self.points[index]
            point.time = datetime.utcnow()

            if self._current_gpx_index >= 0:
                # previous = self.points[self._current_gpx_index]
                point.speed = 10  # todo
            else:
                point.speed = 0

            log.info("generate point %s", point)
            await self.addPoint(point)
            self._current_gpx_index = index
        else:
            log.warning("gpx end track point index=%d", index)

    async def generate(self, interval=1):
        """
        leggiamo il gpx dal file, per sviluppo
        """
        # questa parte serve solo a sincronizzare la partenza del gpx generato con il grabber della camera,
        # solo per lo sviluppo
        log.info("Started gps.generate task")
        self.load_gpx(self.config.gps)

        log.info(
            "camera_id=0, listening on port %s. Waiting for dcam grabber ...",
            self.config.socket_port_gps,
        )
        _local_udp = await aioudp.open_local_endpoint(port=self.config.socket_port_gps)

        # self.resetTimestamp()

        self.time_valid = True

        while True:
            old_point = None
            old_time = time.time()
            for point in self.gpx.walk(only_points=True):
                await (
                    _local_udp.receive()
                )  # questo arriva ogni secondo, oppure 25 frames se siamo in debug
                point.time = datetime.utcnow()

                point.speed = 0  # come mai non c'è nel gpx ?
                if old_point:
                    point.speed = distance(
                        point.latitude,
                        point.longitude,
                        point.elevation,
                        old_point.latitude,
                        old_point.longitude,
                        old_point.elevation,
                    ) / (time.time() / old_time)

                if (
                    random.random() < self.config.gps_stop
                ):  # simuliamo una fermata il 5% dei punti
                    i = 10
                    point.speed = 0
                    while i > 0:
                        log.info("stopping car for a while")
                        await _local_udp.receive()  # questo arriva ogni secondo, oppure 25 frames se siamo in debug
                        point.time = datetime.utcnow()
                        await self.addPoint(point)
                        await asyncio.sleep(1)
                        i -= 1
                    log.info("starting car")

                await self.addPoint(point)
                old_point = point
                # await asyncio.sleep(interval)

    def checkTimeSync(self):
        if os.path.exists("/var/run/ntpd.sock"):
            log.info("checking chrony state")
            # sono su xavier o pc
            # result = subprocess.run(['ntpctl', '-s', 'status'], stdout=subprocess.PIPE)
            # return len(re.findall('clock synced', result)) > 0

            result = subprocess.run(["chronyc", "tracking"], stdout=subprocess.PIPE)
            return (
                len(re.findall(r"Leap status\s*:\s*Normal", result.stdout.decode())) > 0
            )
        return True  # sono su PC


class ContextManagerCursor(sqlite.Cursor):
    def __enter__(self):
        return self

    def __exit__(self, a, b, c):
        self.close()


# {'class': 'RAW', 'device': '/dev/ttyACM0', 'time': 1724936645, 'nsec': 992000000, 'rawdata': [{'gnssid': 6, 'svid': 18, 'snr': 14, 'obs': 'L1C', 'lli': 2, 'locktime': 0, 'freqid': 4, 'pseudorange': 18443367.229919, 'doppler': -462.562469}]}>
# 2024-08-29 13:04:06,026 [INFO] mev_brain.lib.gpslib <dictwrapper: {'class': 'SKY', 'device': '/dev/ttyACM0', 'time': '2024-08-29T13:04:06.000Z', 'nSat': 45, 'uSat': 0, 'satellites': [<dictwrapper: {'PRN': 3, 'el': 15.0, 'az': 120.0, 'ss': 0.0, 'used': False, 'gnssid': 0, 'svid': 3, 'health': 1}>, <dictwrapper: {'PRN': 4, 'el': 37.0, 'az': 63.0, 'ss': 0.0, 'used': False, 'gnssid': 0, 'svid': 4, 'health': 1}>, <dictwrapper: {'PRN': 6, 'el': 51.0, 'az': 229.0, 'ss': 0.0, 'used': False, 'gnssid': 0, 'svid': 6, 'health': 1}>, <dictwrapper: {'PRN': 7, 'el': 56.0, 'az': 169.0, 'ss': 0.0, 'used': False, 'gnssid': 0, 'svid': 7, 'health': 1}>, <dictwrapper: {'PRN': 9, 'el': 73.0, 'az': 39.0, 'ss': 0.0, 'used': False, 'gnssid': 0, 'svid': 9, 'health': 1}>, <dictwrapper: {'PRN': 11, 'el': 41.0, 'az': 296.0, 'ss': 0.0, 'used': False, 'gnssid': 0, 'svid': 11, 'health': 1}>, <dictwrapper: {'PRN': 16, 'el': 13.0, 'az': 63.0, 'ss': 0.0, 'used': False, 'gnssid': 0, 'svid': 16, 'health': 1}>, <dictwrapper: {'PRN': 20, 'el': 24.0, 'az': 307.0, 'ss': 0.0, 'used': False, 'gnssid': 0, 'svid': 20, 'health': 1}>, <dictwrapper: {'PRN': 26, 'el': 5.0, 'az': 31.0, 'ss': 0.0, 'used': False, 'gnssid': 0, 'svid': 26, 'health': 1}>, <dictwrapper: {'PRN': 29, 'el': 1.0, 'az': 342.0, 'ss': 0.0, 'used': False, 'gnssid': 0, 'svid': 29, 'health': 1}>, <dictwrapper: {'PRN': 30, 'el': 27.0, 'az': 195.0, 'ss': 0.0, 'used': False, 'gnssid': 0, 'svid': 30, 'health': 1}>, <dictwrapper: {'PRN': 36, 'el': 36.0, 'az': 163.0, 'ss': 0.0, 'used': False, 'gnssid': 1, 'svid': 123, 'health': 1}>, <dictwrapper: {'PRN': 40, 'el': 22.0, 'az': 126.0, 'ss': 0.0, 'used': False, 'gnssid': 1, 'svid': 127}>, <dictwrapper: {'PRN': 49, 'el': 37.0, 'az': 188.0, 'ss': 0.0, 'used': False, 'gnssid': 1, 'svid': 136, 'health': 1}>, <dictwrapper: {'PRN': 302, 'el': 5.0, 'az': 26.0, 'ss': 0.0, 'used': False, 'gnssid': 2, 'svid': 2, 'health': 1}>, <dictwrapper: {'PRN': 307, 'el': 40.0, 'az': 62.0, 'ss': 0.0, 'used': False, 'gnssid': 2, 'svid': 7, 'health': 1}>, <dictwrapper: {'PRN': 310, 'el': 10.0, 'az': 328.0, 'ss': 0.0, 'used': False, 'gnssid': 2, 'svid': 10, 'health': 1}>, <dictwrapper: {'PRN': 312, 'el': 29.0, 'az': 314.0, 'ss': 0.0, 'used': False, 'gnssid': 2, 'svid': 12, 'health': 1}>, <dictwrapper: {'PRN': 319, 'el': 34.0, 'az': 257.0, 'ss': 0.0, 'used': False, 'gnssid': 2, 'svid': 19, 'health': 1}>, <dictwrapper: {'PRN': 326, 'el': 28.0, 'az': 182.0, 'ss': 0.0, 'used': False, 'gnssid': 2, 'svid': 26, 'health': 1}>, <dictwrapper: {'PRN': 327, 'el': 10.0, 'az': 124.0, 'ss': 0.0, 'used': False, 'gnssid': 2, 'svid': 27, 'health': 1}>, <dictwrapper: {'PRN': 329, 'el': 78.0, 'az': 294.0, 'ss': 0.0, 'used': False, 'gnssid': 2, 'svid': 29, 'health': 2}>, <dictwrapper: {'PRN': 330, 'el': 17.0, 'az': 75.0, 'ss': 0.0, 'used': False, 'gnssid': 2, 'svid': 30, 'health': 1}>, <dictwrapper: {'PRN': 333, 'el': 57.0, 'az': 242.0, 'ss': 0.0, 'used': False, 'gnssid': 2, 'svid': 33, 'health': 1}>, <dictwrapper: {'PRN': 402, 'el': 3.0, 'az': 102.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 2, 'health': 1}>, <dictwrapper: {'PRN': 405, 'el': 21.0, 'az': 123.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 5, 'health': 1}>, <dictwrapper: {'PRN': 406, 'el': 15.0, 'az': 43.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 6, 'health': 1}>, <dictwrapper: {'PRN': 409, 'el': 31.0, 'az': 48.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 9, 'health': 1}>, <dictwrapper: {'PRN': 412, 'el': 1.0, 'az': 342.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 12, 'health': 1}>, <dictwrapper: {'PRN': 416, 'el': 10.0, 'az': 41.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 16, 'health': 1}>, <dictwrapper: {'PRN': 419, 'el': 54.0, 'az': 298.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 19, 'health': 1}>, <dictwrapper: {'PRN': 420, 'el': 66.0, 'az': 66.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 20, 'health': 1}>, <dictwrapper: {'PRN': 422, 'el': 5.0, 'az': 282.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 22, 'health': 1}>, <dictwrapper: {'PRN': 423, 'el': 13.0, 'az': 38.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 23, 'health': 1}>, <dictwrapper: {'PRN': 429, 'el': 13.0, 'az': 219.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 29, 'health': 1}>, <dictwrapper: {'PRN': 435, 'el': 25.0, 'az': 277.0, 'ss': 0.0, 'used': False, 'gnssid': 3, 'svid': 35, 'health': 1}>, <dictwrapper: {'PRN': 65, 'el': 27.0, 'az': 75.0, 'ss': 0.0, 'used': False, 'gnssid': 6, 'svid': 1, 'health': 1}>, <dictwrapper: {'PRN': 66, 'el': 63.0, 'az': 16.0, 'ss': 20.0, 'used': False, 'gnssid': 6, 'svid': 2, 'health': 1}>, <dictwrapper: {'PRN': 67, 'el': 35.0, 'az': 293.0, 'ss': 0.0, 'used': False, 'gnssid': 6, 'svid': 3, 'health': 1}>, <dictwrapper: {'PRN': 75, 'el': 27.0, 'az': 40.0, 'ss': 0.0, 'used': False, 'gnssid': 6, 'svid': 11, 'health': 1}>, <dictwrapper: {'PRN': 76, 'el': 38.0, 'az': 112.0, 'ss': 0.0, 'used': False, 'gnssid': 6, 'svid': 12, 'health': 1}>, <dictwrapper: {'PRN': 77, 'el': 14.0, 'az': 155.0, 'ss': 0.0, 'used': False, 'gnssid': 6, 'svid': 13, 'health': 1}>, <dictwrapper: {'PRN': 81, 'el': 30.0, 'az': 228.0, 'ss': 0.0, 'used': False, 'gnssid': 6, 'svid': 17, 'health': 1}>, <dictwrapper: {'PRN': 82, 'el': 41.0, 'az': 288.0, 'ss': 14.0, 'used': False, 'gnssid': 6, 'svid': 18, 'health': 1}>, <dictwrapper: {'PRN': 83, 'el': 20.0, 'az': 333.0, 'ss': 0.0, 'used': False, 'gnssid': 6, 'svid': 19, 'health': 1}>]}
