import asyncio
import logging
import math
import os
from datetime import datetime
from enum import IntEnum
from xml.etree import ElementTree
from gpxpy.gpx import GPXTrackPoint

import configargparse
import keytree
import numpy as np
import pkg_resources
import webcolors
from fastkml import kml, styles, KML
from shapely.geometry import LineString, Point, shape


log = logging.getLogger(__name__)

parser = configargparse.get_argument_parser()
parser.add_argument(
    "--zones", default="", help="kml with zones"
)
parser.add_argument("--kml-path", default="/tmp", help="kml file to save")
parser.add_argument(
    "--kml-autosave", default=0, type=int, help="gpx autosave frequency in seconds"
)
parser.add_argument(
    "--force-zone", type=int, default=0, help="if 1 force ZoneType.right. Default to 0"
)


class ZoneType(IntEnum):
    off = 0
    right = 1
    both = 2
    vertical_parking = 4


# COLORS = {
#     ZoneType.off: 'red',
#     ZoneType.right: 'blue',
#     ZoneType.both: 'yellow'
#     ZoneType.vertical_parking: 'celeste'
# }


class KmlZones:
    def __init__(self, config):
        self.config = config

        if config.zones:
            self.zones = self.load()

    def __repr__(self):
        r = []
        for name in self.zones:
            zone = self.zones[name]
            r.append(f"""<zone name="{name}" type="{zone['type']}">""")
        return "\n".join(r)

    def load(self):
        tree = ElementTree.fromstring(open(self.config.zones).read())
        kmlns = tree.tag.split("}")[0][1:]

        zones = {}

        root = tree.find(".//{%s}Folder" % kmlns)
        for zone in root.findall(".//{%s}Folder" % kmlns):
            name = zone[0].text
            log.info("loading kml zone name=%s", name)
            zones[name] = dict(name=name, areas=[], type=ZoneType.off)
            if name == "off":
                zones[name]["type"] = ZoneType.off
            if name == "both":
                zones[name]["type"] = ZoneType.both
            elif name == "right":
                zones[name]["type"] = ZoneType.right
            elif name == "vertical":
                zones[name]["type"] = ZoneType.vertical_parking

            for area in zone.findall(".//{%s}Placemark" % kmlns):
                area_name = area[0].text
                for polygon in area.findall(".//{%s}Polygon" % kmlns):
                    # print(polygon)
                    geometry = keytree.geometry(polygon)

                    zones[name]["areas"].append(
                        dict(shape=shape(geometry), name=area_name)
                    )
        return zones

    def check(self, latitude: float, longtitude: float) -> int:  # ZoneType:
        """
        NO priorità delle zone: off, both, right,

        """
        point = Point(longtitude, latitude)

        zones = 0
        # types_matched = []
        for zone in self.zones.values():
            for area in zone["areas"]:
                # print("off point is within ", area['name'], ':',
                if area["shape"].contains(point):
                    # types_matched.append(zone['type'])
                    if zone["type"] == 0:  # prevale su tutti
                        return 0
                    zones += zone["type"]

        # priorità delle zone
        # if ZoneType.off in types_matched:
        #     return ZoneType.off
        # if ZoneType.both in types_matched:
        #     return ZoneType.both
        # if ZoneType.right in types_matched:
        #     return ZoneType.right
        zones += self.config.force_zone

        return zones  # ZoneType.off


def colorname2bgr(name):
    rgb = webcolors.name_to_hex(name)[1:]
    return rgb[4:] + rgb[2:4] + rgb[:2]


class Kml:
    def __init__(self, config, filename=""):
        self.config = config

        self.k = kml.KML()
        self.ns = "{http://www.opengis.net/kml/2.2}"
        self.lineStyle = styles.Style(
            styles=[styles.LineStyle(self.ns, color="77ffff00", width=5)]
        )

        self.d = kml.Document(self.ns, "docid", "track", "doc description")
        self.k.append(self.d)

        self.folder = kml.Folder(self.ns, "fid", "giorno xyz", "f description")
        self.d.append(self.folder)
        self.images_path = pkg_resources.resource_filename("mev_brain", "images")
        self.filename = os.path.join(
            self.config.kml_path,
            filename or datetime.now().strftime("%Y-%m-%d_%H:%M:%S") + ".kml",
        )
        if not os.path.exists(self.config.kml_path):
            os.makedirs(self.config.kml_path)

        self.zones = KmlZones(config)

    def initialize(self):
        loop = asyncio.get_event_loop()
        loop.create_task(self.autoSave())

    def getIconStyle(self, heading=0, color="red", opacity=0.8, object="track"):
        if object == "car":
            # ci sono track da 0 (nord) a 15 (circa 345 gradi)
            icon_href = (
                "http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png"
            )
        else:
            icon_href = "https://earth.google.com/images/kml-icons/track-directional/track-0.png"  # "http://maps.google.com/mapfiles/kml/shapes/cabs.png"

        return styles.Style(
            styles=[
                styles.IconStyle(
                    ns=self.ns,
                    scale=1.0,
                    color="%02x" % int(opacity * 255) + colorname2bgr(color),
                    heading=heading,
                    icon_href=icon_href,
                    # icon_href=f"{self.images_path}/{object}.png",
                )
            ]
        )

    def getHeading(self, point1, point2):
        lon1, lat1 = point1
        lon2, lat2 = point2
        dLon = lon2 - lon1
        y = math.sin(dLon) * math.cos(lat2)
        x = math.cos(lat1) * math.sin(lat2) - math.sin(lat1) * math.cos(
            lat2
        ) * math.cos(dLon)
        brng = np.rad2deg(math.atan2(y, x))
        if brng < 0:
            brng += 360
        return brng

    def getZones(self, location) -> int:
        return self.zones.check(location.latitude, location.longitude)

    async def addRoutePoint(
        self, p1: GPXTrackPoint = None, p2: GPXTrackPoint = None, name="p", color="red"
    ):
        if p2.longitude == 0:
            return

        zones = self.zones.check(p2.latitude, p2.longitude)
        color = "green" if zones > 0 else "black"

        line = kml.Placemark(
            self.ns, "id", name, "description", styles=[self.lineStyle]
        )

        p2_arr = (p2.longitude, p2.latitude)

        heading = 0
        if p1:
            p1_arr = (p1.longitude, p1.latitude)
            line.geometry = LineString([p1_arr, p2_arr])
            self.folder.append(line)
            heading = self.getHeading(p1_arr, p2_arr)

        icon = kml.Placemark(
            self.ns,
            "id",
            name,
            "description",
            styles=[self.getIconStyle(color=color, heading=heading)],
        )
        icon.geometry = Point(p2_arr)
        self.folder.append(icon)

    def addPlate(self, latitude, longitude, name="p", color="red"):
        """usata per debug"""
        zones = self.zones.check(latitude, longitude)
        color = "green" if zones > 0 else "black"

        icon = kml.Placemark(
            self.ns,
            "id",
            name,
            "description",
            styles=[self.getIconStyle(color=color, object="car")],
        )
        icon.geometry = Point((longitude, latitude))
        self.folder.append(icon)

    def save(self, filename=""):
        if not filename:
            filename = self.filename

        with open(filename, "w") as f:
            f.write(self.k.to_string(prettyprint=True))
            log.warning("saving kml to %s", filename)

    async def autoSave(self):
        if not self.config.kml_autosave:
            return

        while True:
            self.save()
            await asyncio.sleep(self.config.kml_autosave)


def exportKml():
    # global parser
    from dnnutils.gpslib import Gps

    parser.add_argument("--export", required=True, help="kml file to save")
    config = parser.parse_args()

    gps = Gps(config)  # , no_save=True)
    kml_out = KML(config)
    kml_zones = KmlZones(config)

    i = 1
    point1 = None
    for point2 in gps.getPoints():
        if point1:
            zones = kml_zones.check(point2.latitude, point2.longitude)
            color = "green" if zones > 0 else "black"
            if zones > 0:
                # point2.name = "%d %s" % (i, str(zone_type).split('.')[1])
                point2.name = str(i)
            kml_out.addRoutePoint(
                (point1.longitude, point1.latitude),
                (point2.longitude, point2.latitude),
                name=point2.name,
                color=color,
            )
            i += 1
        point1 = point2
    kml_out.save(config.export)


if __name__ == "__main__":
    k = Kml()
    p1 = GPXTrackPoint(latitude=45.492734, longitude=11.107233)
    p2 = GPXTrackPoint(latitude=45.492784, longitude=11.107226)
    p3 = GPXTrackPoint(latitude=45.492829, longitude=11.10722)
    k.addRoutePoint(p1, p2, name="1", color="blue")

    k.addRoutePoint(p2, p3, name="2", color="yellow")

    k.addPlate(11.10720, 45.49292, name="AY300AR")
    k.save("/tmp/test.kml")
