import asyncio
import logging
from enum import Enum

import aiosnmp
from mev_common.models import LteStatus

log = logging.getLogger(__name__)


class LteSignal(str, Enum):
    unknown = "unknown"
    bad = "bad"
    sufficient = "sufficient"
    good = "good"
    excellent = "excellent"


class Snmp:
    def __init__(self, ip="127.0.0.1"):
        self.ip = ip

    async def get_lte_signal(self) -> LteStatus:
        try:
            async with aiosnmp.Snmp(host=self.ip, port=161, community="public") as snmp:
                for res in await snmp.bulk_walk("1.3.6.1.4.1.14988.1.1.16.1.1"):
                    #print(res.oid, res.value)
                    lte_param = res.oid.split(".")[
                        -2
                    ]  # "1.3.6.1.4.1.14988.1.1.16.1.1.<lte_param>.x"
                    match lte_param:
                        case "2":
                            self.rssi = (
                                res.value  # "1.3.6.1.4.1.14988.1.1.16.1.1.2.x",  # rssi
                            )
                        case "3":
                            self.rsrq = (
                                res.value  # "1.3.6.1.4.1.14988.1.1.16.1.1.3.4",  # rsrq
                            )
                        case "4":
                            self.rsrp = res.value  # "1.3.6.1.4.1.14988.1.1.16.1.1.4.4",
                        case "7":
                            self.sinr = res.value  # "1.3.6.1.4.1.14988.1.1.16.1.1.7.4",
        except Exception:
            log.warning("snmp error")
            return LteStatus()

        log.debug(
            f"rssi={self.rssi} rsrp={self.rsrp} rsrq={self.rsrq} sinr={self.sinr}"
        )

        lte_status = LteStatus(sinr=self.sinr)

        if self.sinr >= 20:
            lte_status.status = "excellent"
        elif self.sinr >= 13:
            lte_status.status = "good"
        elif self.sinr > 3:
            lte_status.status = "sufficient"
        else:
            lte_status.status = "bad"
        return lte_status


async def test():
    s = Snmp(ip="10.1.109.168")
    status = await s.get_lte_signal()
    print(status)

if __name__ == "__main__":
    asyncio.run(test())
