import asyncio
import json
import logging
import time
from enum import IntEnum
from pathlib import Path
from typing import List

import asset
import configargparse
import cv2 as cv
from autobahn.asyncio.wamp import ApplicationRunner, ApplicationSession
from autobahn.wamp.types import ComponentConfig
from gpxpy.geo import Location
from mev_common.models import (
    DeviceStatus,
    ErrorStatus,
    GpsStatus,
    LoginData,
    OperationStatus,
    Package,
    PlateDb,
    QueryStatus,
)
from pydantic import validate_call

parser = configargparse.get_argument_parser(
    config_file_parser_class=configargparse.YAMLConfigFileParser,
    formatter_class=configargparse.ArgumentDefaultsHelpFormatter,
)


from .db import Db  # noqa: E402
from .lib import gpslib, sensors, truth, utils  # noqa: E402
from .lib.frames import Frames  # noqa: E402
from .lib.kmllib import Kml, ZoneType  # noqa: E402
from .lib.mysnmp import Snmp  # noqa: E402
from .lib.rest import RestApi  # noqa: E402

log = logging.getLogger(__name__)


class ParkingAngle(IntEnum):
    parallel = 0
    vertical = 1


def initParser():
    global parser
    parser.add_argument(
        "-c", "--config", required=False, is_config_file=True, help="config file path"
    )
    parser.add_argument(
        "--local",
        required=False,
        is_config_file=True,
        help="additional local config file path",
    )
    parser.add_argument(
        "--config-save",
        required=False,
        is_write_out_config_file_arg=True,
        help="config file path",
    )

    parser.add_argument("--debug", action="store_true", default=False, help="debug")
    parser.add_argument("--incoming", default="/tmp/json/*.json")
    parser.add_argument("--ws", default="ws://nexus:8082/ws")
    parser.add_argument("--id-device", default="cerbero", help="device's id")
    parser.add_argument("--realm", default="realm1")
    parser.add_argument(
        "--validate-delay",
        type=int,
        default=5,
        help="delay in second to validate plates from two cameras",
    )
    parser.add_argument(
        "--gps-fix-delay",
        type=int,
        default=36,
        help="delay in second to publish fix gps, default to 36",
    )
    parser.add_argument(
        "--plate-images", default="/tmp/images", help="path to load jpg images plate"
    )
    parser.add_argument(
        "--plates-truth", default="", help="path to load file with plates"
    )
    parser.add_argument(
        "--mev-unit", default="cerbero-dev1", help="unit name"
    )  # da dismettere
    parser.add_argument("--flag-gps-fix", default="/tmp/gps-fix", help="gps is valid")
    parser.add_argument(
        "--db-reset", action="store_true", default=False, help="reset database"
    )
    parser.add_argument("--snmp", default="", help="mikrotik ip")
    parser.add_argument(
        "--score-min",
        default=0.1,
        type=float,
        help="filter plates with greater score thant score-min (default 0.7)",
    )

    config, _ = parser.parse_known_args()

    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s %(message)s",
        level="DEBUG" if config.debug else "INFO",
    )

    return config


class PollerMev(ApplicationSession, RestApi):
    def __init__(self, config):
        ApplicationSession.__init__(self, ComponentConfig(config.realm))
        RestApi.__init__(self, config)
        self.settings = config

        now = int(time.time())
        self.gps = gpslib.Gps(config, filename=f"{now}.gpx")
        self.kml: Kml = Kml(config, filename=f"{now}.kml")
        self.save_status = False

        self.db_plates: Db = Db(config.db, reset=config.db_reset)
        self.truth_plates = truth.Truth(config)
        self.frames = Frames()

        self.gps.initialize()
        self.kml.initialize()
        self._counter_gps = 0
        self.gps_started: float = .0
        self.gps.addCallback(self.location_publish)

        self.device_status = DeviceStatus(
            id_device=config.id_device, sensors=sensors.getStats()
        )
        self.device_status.init_list_packages()
        self.latest_gps_status = GpsStatus()

        runner = ApplicationRunner(config.ws, config.realm, extra={})
        asyncio.get_event_loop().run_until_complete(runner.run(self, start_loop=False))

    async def onJoin(self, details):
        log.info("wamp session attached to realm %s", self.settings.realm)

        self.subscribe(self.on_plate, "plate.new")  # from yolodeep
        self.subscribe(self.on_frame, "frame.new")  # from yolodeep
        self.subscribe(self.update_query_status, "query.status")  # from mev-query
        if self.settings.gps == "remote":
            log.info("gps remote: subscribe to gps.remote")
            self.subscribe(self.onGpsRemote, "gps.remote")  # from smartphone
        self.register(self.get_plate_image, "plate.image.get")  # from angular and titan
        self.register(self.getZones, "zones.get")  # from angular
        self.register(self.getStatus, "dcam.status")  # from angular
        self.subscribe(self.onSaveToggle, "dcam.toggle")  # from angular
        self.register(self.config_set, "config.set")  # from titan
        self.subscribe(self.on_yolodeep_start, "yolodeep.start")  # from yolodeep
        self.subscribe(self.on_device_packages, "device.packages")  # from mev-monitor

        asyncio.create_task(self.poll())
        asyncio.create_task(self.status_publish())
        if self.settings.snmp:
            asyncio.create_task(self.poll_snmp())
            # coro = asyncio.to_thread(self.poll_snmp)
            # asyncio.create_task(coro)

    @validate_call
    async def config_set(self, login_data: LoginData):
        try:
            await self.call("query.setlogin", login_data=login_data.dict())
        except Exception:
            return OperationStatus(
                status=False, error=ErrorStatus(message="error in query service")
            ).dict()
        self.settings.id_device = login_data.id_device
        self.device_status.id_device = login_data.id_device
        self.device_status.configured = True
        log.info("setting id_device=%s", self.settings.id_device)
        return OperationStatus(status=True).dict()

    async def poll(self):
        log.info("starting poll plates")
        while True:
            log.debug("gps fix=%s", self.gps.fix)
            if self.gps.fix:
                if not self.gps_started:
                    self.gps_started = time.time()
                    log.info(
                        "gps is valid, now notify systemd to start dcam with flag %s",
                        self.settings.flag_gps_fix,
                    )
                    with open(
                        self.settings.flag_gps_fix, "w"
                    ) as flag:  # watched by systemctl to start dcam.service
                        flag.write("fix")
            else:
                self.latest_gps_status.fixed = False
                self.publish("gps.current", status=self.latest_gps_status.dict())
            await asyncio.sleep(1)
            await self.localizePlates()
            await self.filter_plates(validate_delay=self.settings.validate_delay)
            self.device_status.yolodeep_status.whatchdog()

    async def poll_snmp(self):
        log.info("starting poll snmp")
        self.snmp = Snmp(ip=self.settings.snmp)
        while True:
            self.device_status.lte_status = await self.snmp.get_lte_signal()
            await asyncio.sleep(5)

    @validate_call
    async def update_query_status(self, status: QueryStatus):
        self.device_status.query_status = status

    async def status_publish(self):
        """
        aggiorniamo lo stato della macchina (cerbero) ogni 3 secondi
        """
        while True:
            self.device_status.update()
            self.device_status.sensors = sensors.getStats()
            # log.debug("status device: %s", self.device_status)
            self.publish("device.status", status=self.device_status.dict())
            await asyncio.sleep(3)

    async def location_publish(self, p2=None, **kwargs):
        """
        callback della libreria gpslib
        """
        if self.gps.fix:
            self._counter_gps += 1

            self.device_status.gps_status = GpsStatus(
                latitude=p2.latitude,
                longitude=p2.longitude,
                speed=p2.speed,
                satellites=p2.satellites,
                hdop=p2.horizontal_dilution,
                elevation=p2.elevation,
                fixed=self.gps_started > 0 and (time.time() - self.gps_started) > self.settings.gps_fix_delay, 
            )
            self.latest_gps_status = self.device_status.gps_status
            self.publish("gps.current", status=self.device_status.gps_status.dict())
        # else:
        #     self.device_status.gps_status.fixed = False

    def close(self):
        log.info("closing database")
        self.db_plates.close()
        self.gps.close()

    async def localizePlates(
        self, state_initial=0, state_final=1, state_no_localized=-3, max_time_seconds=10
    ):
        """
        setta latitudine e longitudine dei plates in stato 0
        """

        for plate in self.db_plates.get(state=state_initial):
            # last = self.db_plates.getLatest()
            # for i in range(0, 19):
            #     log.info("last plate " + str(last[i]))
            location = await self.gps.getLocationFromTime(
                timestamp=plate["timestamp"], quarter=plate["quarter"]
            )
            if location:
                self.publish(
                    "gps.marker",
                    location=dict(
                        latitude=location.latitude, longitude=location.longitude
                    ),
                )
                log.info(
                    "plate localized: ascii=%s quarter=%s location=%s",
                    plate["plate_ascii"],
                    plate["quarter"],
                    location,
                )
                self.db_plates.updateLocation(plate, location, state_final)
            else:
                time_elapsed = time.time() - plate["timestamp"]
                delete_plate = time_elapsed > max_time_seconds
                log.warning(
                    "no location for plate ascii=%s timestamp=%f quarter=%f elapsed=%s[sec] deleted=%s",
                    plate["plate_ascii"],
                    plate["timestamp"],
                    plate["quarter"],
                    int(time_elapsed),
                    delete_plate,
                )
                if delete_plate:
                    self.db_plates.updateState(plate, state_no_localized)

    def updateZone(self, plate):
        zones = self.kml.getZones(Location(plate["latitude"], plate["longitude"]))
        self.db_plates.updateZone(plate, zones)
        plate["zone"] = zones
        return zones

    def checkPlateInZone(self, plate, zones):
        """
        torna True se la targa è dentro una zona valida
        """

        if zones == 0:
            return False

        # escludiamo questi falsi positivi di macchine che stanno uscendo da un incrocio o un garage
        if plate["parking"] == ParkingAngle.vertical and not (
            zones & ZoneType.vertical_parking
        ):  # un plate verticale in una zona diversa da vertical_parking non è valido
            return False

        if zones & ZoneType.both:
            return True

        if (zones & ZoneType.right) and plate["quarter"] <= 1:
            return True

        return False

    async def filter_plates(
        self, max_distance=300, validate_delay=10, state_initial=1, state_final=2
    ):
        """
        - elimina le targhe doppie riconosciute da dcam0 e dcam1 tenendo le migliori.
         considera le targhe con un timestamp più vecchio di now di almeno validate_delay
        - localizza latitudine e longitudine nella zona
        """
        processed_plates = []
        for plate in self.db_plates.getOlder(
            state=state_initial, validate_delay=validate_delay
        ):
            if plate["key"] in processed_plates:
                continue

            similar_plates = self.db_plates.getSimilar(
                plate["plate_ascii"], max_distance=max_distance, state=state_initial
            )
            best = True
            for p in similar_plates:
                if best:
                    best = False  # solo la prima ha lo score maggiore
                    plate_best = PlateDb(**dict(p))
                    if plate_best.score >= self.settings.score_min:
                        state_for_best = state_final
                        plate_best.state = state_for_best
                        self.db_plates.updateState(p, state_for_best)
                        self.publish(
                            "plate.valid", plate=plate_best.dict()
                        )  # plate_best)
                        log.info("plate.valid: %s", plate_best)
                        if self.settings.plates_truth:
                            self.truth_plates.add(plate_best.plate_ascii)
                    else:
                        log.info(
                            "plate rejected with low score than %s: %s",
                            self.settings.score_min,
                            plate_best,
                        )
                        self.db_plates.updateState(p, -state_final)  # scartiamo
                else:
                    # scartiamo questa targa
                    self.db_plates.updateState(p, -state_final)

                processed_plates.append(p["key"])

    async def saveToggle(self, status):
        log.info("saveToggle(%s)", status)
        if status:
            utils.sendSignalToProcesses(
                "dcam_run", sig="SIGUSR1"
            )  # salviamo file video mandando un segnale SIGUSR1 ai processi dcam_run
            await self.gps.track()
        else:
            utils.sendSignalToProcesses(
                "dcam_run", sig="SIGUSR2"
            )  # interrompiamo la registrazione
            await self.gps.stop()
        self.save_status = status

    def getStatusSaving(self):
        return self.save_status

    # @Authorize()
    # async def getContextImage(self, request, audience="", payload={}, key=""):
    #     resp = await self.getPlateImage(
    #         key=key, size=None, for_http=False, add_crop=True
    #     )
    #     return response.json(resp)

    async def onDisconnect(self):
        log.warning("wamp transport lost, shutting down")
        self.close()
        asyncio.get_event_loop().stop()

    async def on_frame(self, source_id: int = 0, frame_num: int = 0, path: str = ""):
        self.frames.add(source_id, frame_num, path)
        self.device_status.yolodeep_status.update(source_id, frame_num)
        if self.settings.gps == "dummy":
            await self.gps.generate_point()

    async def on_plate(self, plate=None):
        """
        se il GPS time non è corretto non inseriamo le targhe
        """
        if not self.gps.time_valid:
            log.warning("GPS time not ready, skip new plate detected")
            return
        # if not ("DK" in plate['plate_ascii']):
        #    return
        self.db_plates.add(**plate)

    async def onSaveToggle(self, status=False):
        await self.saveToggle(status)

    async def onGpsRemote(self, *args, **kwargs):
        log.info("got %s", kwargs)
        await self.gps.addPointFromRemote(kwargs["coords"], kwargs["timestamp"])

    async def getStatus(self, *args, **kwargs):
        return {"saving": self.getStatusSaving()}

    # @Authorize()
    # async def getImage(
    #     self, request, audience="", payload={}, key="", kind="context", encode="binary"
    # ):
    #     """
    #     TODO encode==base64 non implementato
    #     """
    #     if key == "latest":
    #         plate = self.db_plates.getLatest()
    #         if plate:
    #             key = plate["key"]
    #         else:
    #             return response.text("")
    #     _bytes = await self.get_plate_image(
    #         key=key,
    #         size=None,
    #         binary=encode == "binary",
    #         for_http=False,
    #         add_crop=True,
    #         kind=kind,
    #     )
    #     return response.raw(_bytes, headers={"Content-type": "image/jpeg"})

    # async def capture_and_encode_frames(self):
    #     filename = asset.load("mev_brain:assets/cerbero.jpg").filename
    #     with open(filename, mode="rb") as f:
    #         image = f.read()
    #     while True:
    #         yield (b'--frame\r\n'
    #             b'Content-Type: image/jpeg\r\n\r\n' + image + b'\r\n')

    async def get_camera_image(self, request, id: int = 0):
        async def _jpeg(response):
            image, filename = self.frames.get_frame_binary(source_id=id)
            log.debug("stream frame '%s'", filename)
            if image is not None:
                await response.send(image)

        response = await request.respond(content_type="image/jpeg")
        await _jpeg(response)

    async def get_camera_stream(self, request, id: int = 0):
        framerate = int(request.args.get("framerate", 5))

        async def stream_jpeg(response):
            while True:
                image, filename = self.frames.get_frame_binary(source_id=id)
                log.debug("stream frame '%s'", filename)
                if image is not None:
                    await response.send(
                        b"--frame\r\nContent-Type: image/jpeg\r\n\r\n" + image + b"\r\n"
                    )
                await asyncio.sleep(1 / framerate)

        response = await request.respond(
            content_type="multipart/x-mixed-replace; boundary=frame"
        )
        await stream_jpeg(response)

    async def get_image_stream(self, request):
        async def stream_jpeg(response):
            filename = asset.load("mev_brain:assets/cerbero.jpg").filename
            with open(filename, mode="rb") as f:
                image: bytes = f.read()

            i = 0
            while True:
                i += 1
                log.debug("stream frame %i", i)
                await response.send(
                    b"--frame\r\nContent-Type: image/jpeg\r\n\r\n" + image + b"\r\n"
                )
                await asyncio.sleep(0.2)

        response = await request.respond(
            content_type="multipart/x-mixed-replace; boundary=frame"
        )
        await stream_jpeg(response)

    # async def get_image_stream(self, request):
    #     response = await request.respond() #content_type='multipart/x-mixed-replace; boundary=frame')
    #     return response.stream(self.capture_and_encode_frames(), mimetype='multipart/x-mixed-replace; boundary=frame')

    def _resize_plate(self, img, overlay, size_overlay_perc=30):
        h, w = img.shape[:2]
        w_overlay = int(w * size_overlay_perc / 100)
        ratio = float(w_overlay) / overlay.shape[1]
        h_overlay = int(overlay.shape[0] * ratio)
        overlay = cv.resize(
            overlay, (w_overlay, h_overlay), interpolation=cv.INTER_AREA
        )
        return overlay

    @validate_call
    async def on_device_packages(self, packages: List[Package]):
        log.info("on_device_packages: %s", packages)
        self.device_status.set_packages(packages)

    async def on_yolodeep_start(self, plates_truth: str = "", gpx: str = ""):
        if plates_truth:
            self.settings.plates_truth = plates_truth
            self.truth_plates = truth.Truth(self.settings)
        if gpx:
            self.gps.load_gpx(gpx)

    async def get_plate_image(
        self,
        key="",
        size=(100, 50),
        binary=False,
        for_http=True,
        add_crop=True,
        add_license_text=False,
        kind="context",
        size_overlay_perc=30,
    ):
        """
        blabla/{plate.source_id}_{plate.frame_num:07d}_{left}-{top}-{width}-{height}_{plate.nationality}_{plate.license}.jpg
        """

        p = Path(self.settings.plate_images)
        contexts = list(p.glob(f"{key}_*.jpg"))
        if not contexts or len(contexts) > 1:
            log.warning(
                "plate.image.get key=%s len(plates) no single image in path=%s",
                key,
                self.settings.plate_images,
            )
            return ""
        filename = str(contexts[0])
        # try:
        #     filename = str(contexts[0])
        #     match = re.match(
        #         ".*_(?P<x>.*).*-(?P<y>.*).*-(?P<w>.*).*-(?P<h>.*).*_.*_(?P<license>.*)\..*",
        #         filename,
        #     )
        #     mm = match.groupdict()
        #     x = int(mm["x"])
        #     y = int(mm["y"])
        #     w = int(mm["w"])
        #     h = int(mm["h"])

        json_filename = filename.replace(".jpg", ".json")
        try:
            with open(json_filename, "r") as f:
                payload = json.load(f)
                x, y, w, h = payload["plate_box"]
            log.info("get_plate_image key=%s payload=%s", key, payload)
        except Exception:
            log.error("get_plate_image: key=%s cannot parse %s", key, json_filename)
            return ""

        # log.info("mm=%s", mm)

        img = cv.imread(filename)
        if add_crop:
            plate_img = img[y : y + h, x : x + w]

            if size_overlay_perc:
                plate_img = self._resize_plate(
                    img, plate_img, size_overlay_perc=size_overlay_perc
                )
            label = ""
            if add_license_text:
                label = payload["license"]  # mm["license"]
            utils.add_overlay(img, plate_img, label=label, fontSize=7.0, tickness=5)

        if binary:
            return utils.toBinary(img)
        data = utils.getBase64(img, size, for_http=for_http)
        return {"base64": data}

    async def getZones(self):
        if self.settings.zones:
            with open(self.settings.zones) as f:
                return {"kml": f.read()}
        else:
            return {"kml": ""}


def main():
    config = initParser()
    if config.debug:
        import jurigged

        logging.getLogger("watchdog").setLevel(logging.INFO)
        jurigged.watch(pattern="*mev_brain*.py")

    log.info("Starting brain: id-device=%s", config.id_device)
    poller = PollerMev(config)
    try:
        asyncio.get_event_loop().run_forever()
    except KeyboardInterrupt:
        poller.close()
    finally:
        print("Bye")
