import asyncio

import configargparse
from autobahn.asyncio.wamp import ApplicationRunner, ApplicationSession
from mev_common.models import (
    DeviceStatus,
    GpsStatus,
    LoginData,
    OperationStatus,
    ErrorStatus,
    PlateQuery,
    PlateDb,
)
from pydantic import validate_call


def get_config():
    parser = configargparse.get_argument_parser()
    parser.add_argument(
        "--exposure", default=0, type=float, help="exposure time in microseconds"
    )
    parser.add_argument(
        "--brightness", default=0, type=float, help="brightness target from 0.0 to 1.0"
    )
    parser.add_argument("--frame", default=0, type=int, help="save frame")
    parser.add_argument("--step", action="store_true", default=False)
    parser.add_argument("--dry-run", action="store_true", default=False)
    parser.add_argument("--ws", default="ws://nexus:8082")
    parser.add_argument("--realm", default="realm1")

    parser.add_argument("--username", default="", help="test login")
    parser.add_argument("--backend", default="", help="switch to backend")
    parser.add_argument("--plate", default="", help="test plate, example EN675TL")
    parser.add_argument("--upgrade", action="store_true", default=False, help="send upgrade signal")
    parser.add_argument("--packages", default="", help="packages to upgrade")
    parser.add_argument(
        "--listen",
        action="store_true",
        default=False,
        help="subscribe for incoming messages",
    )
    parser.add_argument(
        "--docs",
        action="store_true",
        default=False,
        help="generate docs",
    )

    config, _ = parser.parse_known_args()
    return config


class Component(ApplicationSession):
    config = None

    async def onJoin(self, details):
        print(f"connected to {self.config.ws}")

        self._printed = []

        # self.subscribe(self.pipeline.onGps, "gps")
        if self.config.exposure:
            self.publish("camera.exposure", exposure=self.config.exposure)
            self.leave()
        if self.config.brightness:
            self.publish("camera.brightness", brightness=self.config.brightness)
            self.leave()
        if self.config.frame:
            self.publish("frame.save", frame_num=self.config.frame)
            self.leave()
        if self.config.step:
            self.publish("pipeline.step")
            self.leave()
        if self.config.username:
            login_data = LoginData(kc_username=self.config.username)
            result = await self.call("config.set", login_data=login_data.dict())
            print(f"config.set: {result}")
            self.leave()
        if self.config.backend:
            result = await self.call("query.setbackend", backend=self.config.backend)
            print(f"query.setbackend: {result}")
            self.leave()
        if self.config.plate:
            self.publish(
                "plate.valid",
                plate=PlateDb(plate=self.config.plate, nationality="ITA").dict(),
            )
            self.leave()
        if self.config.upgrade:
            self.publish("device.upgrade.start", packages=self.config.packages.split(","), dry_run=self.config.dry_run)
        if self.config.listen:
            print("listen: device.status, device.upgrade*, yolodeep.start")
            self.subscribe(self.on_device_status, "device.status")
            self.subscribe(self.on_device_upgrade, "device.upgrade")
            self.subscribe(self.on_device_upgrade, "device.upgrade.end")
            self.subscribe(self.on_yolodeep_start, "yolodeep.start")
        if self.config.docs:
            print("## RPC\n")
            self._print(
                "rpc=**config.set**",
                LoginData(kc_client_secret="xxx"),
                desc="Set configuration",
                result=f"""
positive result
```javascript
{OperationStatus(status=True).model_dump_json(indent=2)}
```
negative result
```javascript
{OperationStatus(status=False, error=ErrorStatus()).model_dump_json(indent=2)}
```
""",
            )
            print("## Topics\n")
            self.subscribe(self.on_device_status_docs, "device.status")
            self.subscribe(self.on_gps_current_docs, "gps.current")
            self.subscribe(self.on_plate_query_docs, "plate.query")
            self.subscribe(self.on_plate_docs, "plate.valid")
            self.subscribe(self.on_device_upgrade, "device.upgrade")

    def onLeave(self, details):
        self.disconnect()

    def onDisconnect(self):
        asyncio.get_event_loop().stop()

    @validate_call
    async def on_device_status(self, status: DeviceStatus):
        print("device.status:", status.dict())

    def _check_if_not_printed(self, what):
        if what in self._printed:
            return False
        self._printed.append(what)
        return True

    def _print(self, title, obj, desc="", result=""):
        print(f"### {title}")
        print()
        if desc:
            print(desc)
            print()
        print("```javascript")
        print(obj.model_dump_json(indent=2))
        print("```")
        print()
        if result:
            print(result)
            print()

    @validate_call
    async def on_device_status_docs(self, status: DeviceStatus):
        if self._check_if_not_printed("device.status"):
            self._print(
                "topic=**device.status**",
                status,
                desc="device status raised every 3 seconds",
            )

    @validate_call
    async def on_gps_current_docs(self, status: GpsStatus):
        if self._check_if_not_printed("gps.current"):
            self._print(
                "topic=**gps.current**", status, desc="gps location raised every second"
            )

    @validate_call
    async def on_plate_query_docs(self, plate: PlateQuery):
        if self._check_if_not_printed("plate.query"):
            self._print(
                "topic=**plate.query**",
                plate,
                desc="plate query response from cerbero BE",
            )

    @validate_call
    async def on_plate_docs(self, plate: PlateDb):
        if self._check_if_not_printed("plate.valid"):
            self._print("topic=**plate.valid**", plate, desc="plate detected from ocr")

    async def on_device_upgrade_docs(self, *args, **kwargs):
        self._print("topic=**device.upgrade**", kwargs, desc="device upgrade")

    async def on_device_upgrade(self, *args, **kwargs):
        print(f"on_device_upgrade: kwargs={kwargs}")
    
    async def on_yolodeep_start(self, *args, **kwargs):
        print("on_yolodeep_start: kwargs={kwargs}")


def run():
    config = get_config()
    runner = ApplicationRunner(
        config.ws,
        config.realm,
    )
    Component.config = config
    try:
        runner.run(Component)
    except Exception as e:
        print("error %s", e)


if __name__ == "__main__":
    run()
