import importlib
import logging
import time
from typing import List, Optional

import psutil
from pydantic import BaseModel

log = logging.getLogger(__name__)

PACKAGES = [
    "mev-cerbero",
    "mev-monitor",
    "mev-assets",
    "mev-frontend",
    "mev-yolodeep",
    "mev-fake",
]


class LoginData(BaseModel):
    kc_username: str = "bn.ufficio"
    kc_pass: str = ""  # lasciarlo vuoto
    kc_url: str = "https://mobile-test.cloudtitan.it/auth/realms/titan-up-realm/protocol/openid-connect/token"
    kc_client_id: str = "tup-web-login-client"
    kc_client_secret: str = "fCq5qsDRKfJtGZBqxnGo2lhhH56Htelz"
    kc_grant_type: str = "password"
    cerbero_be_url: str = "https://mobile-test.cloudtitan.it/cerbero/rest/v1/cruscotto"
    id_cliente: str = "PROVA"
    id_device: str = "cerbero-dev0"


# class ConfigData(BaseModel):
#     id_device: str = "cerbero-dev0"


class Plate(BaseModel):
    plate: str = "EN675TL"
    nationality: str = "ITA"
    key: int = 0


class PlateDb(BaseModel):
    """
    {
    "id": 64,
    "key": "0_0000169",
    "object_id": "73",
    "state": 2,
    "plate_ascii": "DN188AF",
    "plate": "DN188AF",
    "score": 0.9342857142857143,
    "timestamp": 1714119629.904649,
    "quarter": 0,
    "x": 846,
    "y": 789,
    "w": 436,
    "h": 287,
    "angle": -29.8,
    "parking": 1,
    "latitude": 45.43218295352518,
    "longitude": 10.967575659025803,
    "elevation": 59.412669240917836,
    "speed": 5.519103195386955,
    "zone": 0,
    "nationality": "ITA",
    "distance": 0
    }
    """

    id: int = 1
    key: str = ""
    object_id: str = "1"
    state: int = 0
    plate_ascii: str = ""
    plate: str = ""
    score: float = 0.0
    timestamp: float = 0.0
    quarter: int = 0
    x: int = 1
    y: int = 1
    w: int = 10
    h: int = 10
    angle: float = 0.0
    parking: int = 1
    latitude: float = 0.0
    longitude: float = 0.0
    elevation: float = 0.0
    speed: float = 0.0
    zone: int = 0
    nationality: str = ""
    distance: int = 0


class PlateQuery(BaseModel):
    plate: str = "EN675TL"
    lat: float = 0.0
    lon: float = 0.0
    key: str = ""
    query: dict = {}
    latency: float = 0


class QueryStatus(BaseModel):
    logged: bool = False
    latency: float = .0
    up: bool = False
    total: int = 0
    errors: int = 0
    no_insurance: int = 0
    no_mot: int = 0


class GpsStatus(BaseModel):
    latitude: float = 0.0
    longitude: float = 0.0
    satellites: int = 0
    speed: float = 0.0
    hdop: float = 0.0
    elevation: float = 0.0
    fixed: bool = False


class Package(BaseModel):
    name: str = "mev-cerbero"
    version: str = "0.0.0"
    latest_version: str = ""


class LteStatus(BaseModel):
    # connected: bool = False
    sinr: int = 0
    status: str = "unknown"


class YolodeepStatus(BaseModel):
    up: bool = False
    error: str = ""
    time: float = 0.0
    frame_num: dict = {0: 0, 1: 0}

    def update(self, source_id, frame_num):
        self.time = time.time()
        self.up = True
        self.frame_num[source_id] = frame_num

    def whatchdog(self, timeout=5):
        if time.time() - self.time > timeout:
            self.up = False


class DeviceStatus(BaseModel):
    id_device: str = "cerbero-dev0"
    sensors: dict = {}  # {"temps": {"CPU": 40.0}}
    query_status: QueryStatus = QueryStatus()
    gps_status: GpsStatus = GpsStatus()
    lte_status: LteStatus = LteStatus()
    yolodeep_status: YolodeepStatus = YolodeepStatus()
    time: float = 0.0
    uptime: float = 0.0
    packages: List[Package] = []  # aggiornata poi da mev-monitor
    configured: bool = False
    need_upgrade: bool = False
    session: int = 0  # usata per identificare una sessione di lavoro

    def update(self):
        self.time = time.time()
        self.uptime = self.time - psutil.boot_time()

    def set_packages(self, packages=List[Package]):
        if packages:
            self.packages = packages
            self.need_upgrade = False
            for package in self.packages:
                if package.latest_version and package.version != package.latest_version:
                    self.need_upgrade = True

    def init_list_packages(self):
        self.update()
        self.packages = []
        for package in PACKAGES:
            try:
                version = importlib.metadata.version(package)
                self.packages.append(Package(name=package, version=version))
            except Exception:
                log.warning("cannot get package '%s' version", package)


class ErrorStatus(BaseModel):
    message: str = "error message"
    code: int = 500


class OperationStatus(BaseModel):
    status: bool = False
    error: Optional[ErrorStatus] | None = None


if __name__ == "__main__":
    print("LoginData:", LoginData().model_dump_json(indent=2))
    print("DeviceStatus:", DeviceStatus().model_dump_json(indent=2))
