import logging
from typing import Any, Optional

import aiohttp
import numpy as np
import yaml
from mev_common.models import LoginData
from pydantic import BaseModel

log = logging.getLogger(__name__)


class Visura(BaseModel):
    keyPlate: Optional[str] = ""
    tipoVeicolo: Optional[str] = ""
    targa: Optional[str] = ""
    assicurato: Optional[bool] = True
    revisionato: Optional[bool] = True
    requireCheck: Optional[bool] = False
    esitoVisura: Optional[Any] = None
    error: Optional[bool] = False
    errorDescr: Optional[Any] = None


class Indirizzo(BaseModel):
    keyPlate: Optional[str] = ""
    numCivico: Optional[str] = ""
    strada: Optional[str] = ""
    cap: Optional[str] = ""
    citta: Optional[str] = ""
    provincia: Optional[str] = ""
    regione: Optional[str] = ""
    nazione: Optional[str] = ""


class ResponseQuery(BaseModel):
    visura: Optional[Visura] = None
    indirizzo: Optional[Indirizzo] = None


class Query(BaseModel):
    responseQuery: ResponseQuery
    latency: float = 0.0


class Response(BaseModel):
    plate: str
    key: str
    query: Query


async def on_request_start(session, trace_config_ctx, params):
    log.warning("request: url='%s' headers=%s", params.url, dict(params.headers))


async def on_request_end(session, trace_config_ctx, params):
    log.warning("response: %s", await params.response.text())


class Api:
    def __init__(self, config, debug=False, seed=205, backend="s21"):
        self.main_config = config
        self.trace_config = aiohttp.TraceConfig()
        self.debug = debug
        self.rng = np.random.default_rng(seed)
        self.login = False
        self.timeout = 10 # ridefinito poi
        self.autologin = False
        if self.set_backend(backend):
            self.timeout = aiohttp.ClientTimeout(total=int(self.config["timeout"]))
            self.autologin = self.config.get("autologin", False)
            if "kc_grant_type" not in self.config:
                self.config["kc_grant_type"] = "kc_pass"

            if debug:  # logging.DEBUG >= logging.root.level:
                log.warning("enable debug")
                self.trace_config.on_request_start.append(on_request_start)
                self.trace_config.on_request_end.append(on_request_end)

            self.login: bool = (
                self.hasToken()
            )  # nel caso il token sia statico nella configurazione
            # log.info("token preloaded: %s", self.login)

    def __repr__(self):
        return str(self.config)

    async def set_login(self, login_data: LoginData):
        self.config["kc_username"] = login_data.kc_username
        if login_data.kc_pass:
            self.config["kc_pass"] = login_data.kc_pass
            self.config["kc_client_id"] = login_data.kc_client_id
            self.config["kc_url"] = login_data.kc_url
            self.config["kc_grant_type"] = login_data.kc_grant_type
            self.config["kc_client_secret"] = login_data.kc_client_secret
            self.config["calls"]["id_cliente"] = login_data.id_cliente
            self.config["calls"]["cerbero_be_url"] = login_data.cerbero_be_url

        logged = await self.get_token()
        self.autologin = True  # da adesso in poi ritenta l'autologin
        return logged

    def set_backend(self, backend: str):
        log.info("set backend %s", backend)
        self.backend = backend
        try:
            self.config = yaml.safe_load(self.main_config.api)[backend]
        except Exception:
            log.error("backend %s not defined", backend)
            return False
        self.autologin = self.config.get("autologin", False)
        return True

    def hasToken(self):
        return self.config.get("token", "") != ""

    def resetToken(self):
        self.config["token"] = ""
        self.login = False

    async def get_token(self):
        """Return the access token"""
        pass

    async def refreshToken(self) -> bool:
        """Refresh the token"""
        pass

    async def base_query(self, plate="EN675TL", simulate_expired=False):
        """Query to backend service"""
        pass

    def _set_token_request(self) -> dict:
        """Set the request to get the access token for the backend service"""
        pass

    def _select_url(self) -> str:
        return self.config["calls"]["cerbero_be_url"]


def _select_mock_url(self) -> str:
    """Fuction to call fake backends"""

    base = self.config["calls"]["cerbero_be_url"]
    options = ["cerbero_be_url", "access-denyed", "timeout"]
    if self.forced_url or self.forced_url not in range(len(options)):
        chosen_url = self.rng.integers(len(options))
    else:
        chosen_url = self.forced_url
    if chosen_url == 0:
        log.debug(f"base url selected => {base}")
        return base
    elif chosen_url == 1:
        log.debug(f"access-denyed url selected => {base}/{options[1]}")
        return base + "/" + options[1]
    elif chosen_url == 2:
        log.debug(f"timeout url selected => {base}/{options[2]}")
        return base + "/" + options[2]
    else:
        return base
