import asyncio

import aiohttp
from mev_common.models import PlateDb


from .common import Api, Indirizzo, Response, ResponseQuery, Query, Visura, log


class S21(Api):
    def __init__(self, config, debug=False, seed=205, backend="s21"):
        super().__init__(config, debug, seed, backend)

    def _get_token_request(self) -> dict:
        payload = {
            "username": self.config["kc_username"],
            "password": self.config["kc_pass"],
            "client_id": self.config["kc_client_id"],
            "grant_type": self.config["kc_grant_type"],
            "client_secret": self.config["kc_client_secret"],
        }
        return payload

    async def get_token(self) -> bool:
        try:
            async with aiohttp.ClientSession(
                timeout=self.timeout, trace_configs=[self.trace_config]
            ) as session:
                request = self._get_token_request()
                log.info(
                    "get_token: username=%s url=%s",
                    request["username"],
                    self.config["kc_url"],
                )
                resp = await session.post(
                    self.config["kc_url"], data=self._get_token_request()
                )
                if resp.status == 200:
                    payload = await resp.json()
                    log.debug("response payload: %s", payload)
                    self.login = True
                    try:
                        self.config["token"] = payload["access_token"]
                        self.config["expire"] = payload["expires_in"]
                        log.info(f"logged in {self.backend}")
                    except Exception as e:
                        log.error(str(e))
                        self.login = False
                        return self.login
                elif resp.status == 400:
                    log.warning(
                        "invalide response code=400 response=%s", await resp.json()
                    )
                    self.login = False
                else:
                    log.warning(
                        "invalide response code=%s: %s", resp.status, await resp.text()
                    )
                    self.login = False
                return self.login

        except (aiohttp.ClientError, asyncio.TimeoutError):
            log.warning(f"connection error to {self.backend}")
            self.login = False
            return self.login

    async def base_query(self, plate: PlateDb) -> ResponseQuery | None:
        """visura S21 backend"""
        params = {
            "targa": plate.plate,
            "idCliente": self.config["calls"]["id_cliente"],
            "lat": plate.latitude,
            "lon": plate.longitude,
        }

        async with aiohttp.ClientSession(
            trace_configs=[self.trace_config],
            timeout=self.timeout,
        ) as session:
            headers = {"Authorization": f"Bearer {self.config['token']}"}
            try:
                chosen_url = self._select_url()
                log.debug(
                    f"http get: backend={self.backend} url={chosen_url} params={params} headers=..."
                )
                resp = await session.get(chosen_url, params=params, headers=headers)
            except asyncio.TimeoutError:
                return None
            except Exception as e:
                log.error(e)
                return None
            if resp.status in (204, 200):  # esito positivo
                payload = await resp.json()
                log.info(f"response from {self.backend}: {payload}")
                if payload and "visura" in payload:
                    visura = payload["visura"]
                    visura.pop("keyPlate", None)
                    indirizzo = payload["indirizzo"]
                    indirizzo.pop("keyPlate", None)
                    response = Response(
                        plate=plate.plate,
                        key=plate.key,
                        query=Query(
                            responseQuery=ResponseQuery(
                                visura=Visura(**visura),
                                indirizzo=Indirizzo(**indirizzo),
                            )
                        ),
                    )
                    return response
                else:
                    log.error("response without payload")
                    return None
            elif resp.status == 401:
                log.warning(
                    "access denyed: %s, try to refresh token", await resp.text()
                )
                self.resetToken()
            log.warning(f"query failed with http code:{resp.status}")
            return None


class FakeS21(S21):
    def __init__(
        self, config, debug=False, seed=205, forced_url=None, backend="fake-s21"
    ):
        super().__init__(config, debug, seed, backend)
        self.forced_url = forced_url
